from enum import Enum


class EnumVMListItemStatus(Enum):
    deploying = "deploying"
    running = "running"
    halted = "halted"
    paused = "paused"
    halting = "halting"
    migrating = "migrating"
    starting = "starting"
    error = "error"



"""
Auto-generated class for Filesystem
"""

from . import client_support


class Filesystem(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(mountpoint, name, parent, quota, readOnly, sizeOnDisk):
        """
        :type mountpoint: str
        :type name: str
        :type parent: str
        :type quota: int
        :type readOnly: bool
        :type sizeOnDisk: int
        :rtype: Filesystem
        """

        return Filesystem(
            mountpoint=mountpoint,
            name=name,
            parent=parent,
            quota=quota,
            readOnly=readOnly,
            sizeOnDisk=sizeOnDisk,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'Filesystem'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'mountpoint'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.mountpoint = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'name'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.name = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'parent'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.parent = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'quota'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.quota = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'readOnly'
        val = data.get(property_name)
        if val is not None:
            datatypes = [bool]
            try:
                self.readOnly = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'sizeOnDisk'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.sizeOnDisk = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



import psutil
from ..healthcheck import HealthCheckRun


descr = """
Check open file descriptors for each node process, if it exceeds 90% of the soft limit, it raises a warning,
if it exceeds 90% of the hard limit, it raises an error.
"""


class OpenFileDescriptor(HealthCheckRun):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('openfile-descriptors', 'Open File Descriptors', 'System Load', resource)
        self.node = node

    def run(self):
        for process in self.node.client.process.list():
            for rlimit in process['rlimit']:
                if rlimit['resource'] == psutil.RLIMIT_NOFILE:
                    pid = str(process['pid'])
                    if (0.9 * rlimit['soft']) <= process['ofd'] < (0.9 * rlimit['hard']):
                        self.add_message(pid, 'WARNING', 'Open file descriptors for process %s exceeded 90%% of the soft limit' % pid)
                    elif process['ofd'] >= (0.9 * rlimit['hard']):
                        self.add_message(pid, 'ERROR', 'Open file descriptors for process %s exceeded 90%% of the hard limit' % pid)
                    break

        if not self._messages:
            self.add_message('-1', 'OK', 'Open file descriptors for all processes are within limit')



def get_configuration_and_service(ays_repo):
    services = ays_repo.servicesFind(actor='configuration')
    if len(services) > 1:
        raise RuntimeError('Multiple configuration services found')

    service = services[0] if services else None
    configuration = service.model.data.to_dict()['configurations'] if service else []

    return {conf['key']: conf['value'] for conf in configuration}, service


def get_configuration(ays_repo):
    configs, _ = get_configuration_and_service(ays_repo)
    return configs


def refresh_jwt_token(token):
    import requests
    import time
    headers = {'Authorization': 'bearer %s' % token}
    params = {'validity': '3600'}
    resp = requests.get('https://itsyou.online/v1/oauth/jwt/refresh', headers=headers, params=params)
    resp.raise_for_status()
    return resp.content.decode()


def get_jwt_token(ays_repo):
    import jose
    import jose.jwt
    import requests
    import time
    configs, service = get_configuration_and_service(ays_repo)
    jwt_token = configs.get('jwt-token', '')
    jwt_key = configs.get('jwt-key')
    if not jwt_token:
        return jwt_token

    try:
        token = jose.jwt.decode(jwt_token, jwt_key)
        if token['exp'] < time.time() - 240:
            jwt_token = refresh_jwt_token(jwt_token)

        if token['exp'] > time.time() + 3600:
            raise RuntimeError('Invalid jwt-token expiration time is too long should be less than 3600 sec')

    except jose.exceptions.ExpiredSignatureError:
        jwt_token = refresh_jwt_token(jwt_token)
    except Exception:
        raise RuntimeError('Invalid jwt-token and jwt-key combination')

    for config in service.model.data.configurations:
        if config.key == 'jwt-token':
            config.value = jwt_token
            break

    service.saveAll()
    return jwt_token


def get_jwt_token_from_job(job):
    if 'token' in job.context:
        return job.context['token']

    return get_jwt_token(job.service.aysrepo)



import io
import time
import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class StorageEngine:
    """storageEngine server"""

    def __init__(self, name, container, bind='0.0.0.0:16379', data_dir='/mnt/data', master=None):
        """
        TODO: write doc string
        """
        self.name = name
        self.master = master
        self.container = container
        self.bind = bind
        self.port = int(bind.split(':')[1])
        self.data_dir = data_dir
        self.master = master
        self._ays = None

    @classmethod
    def from_ays(cls, service, password=None):
        logger.debug("create storageEngine from service (%s)", service)
        from .Container import Container

        container = Container.from_ays(service.parent, password)
        if service.model.data.master != '':
            master_service = service.aysrepo.serviceGet('storage_engine', service.model.data.master)
            master = StorageEngine.from_ays(master_service, password)
        else:
            master = None

        return cls(
            name=service.name,
            container=container,
            bind=service.model.data.bind,
            data_dir=service.model.data.homeDir,
            master=master,
        )

    def _configure(self):
        logger.debug("configure storageEngine")
        buff = io.BytesIO()
        self.container.client.filesystem.download('/etc/ardb.conf', buff)
        content = buff.getvalue().decode()

        # update config
        content = content.replace('/mnt/data', self.data_dir)
        content = content.replace('0.0.0.0:16379', self.bind)

        mgmt_bind = "%s:%s" % (self.container.node.addr, self.port)
        if self.bind != mgmt_bind:
            content += "server[1].listen %s\n" % mgmt_bind

        if self.master is not None:
            _, port = self.master.bind.split(":")
            content = content.replace('#slaveof 127.0.0.1:6379', 'slaveof {}:{}'.format(self.master.container.node.addr, port))

        # make sure home directory exists
        self.container.client.filesystem.mkdir(self.data_dir)

        # upload new config
        self.container.client.filesystem.upload('/etc/ardb.conf.used', io.BytesIO(initial_bytes=content.encode()))

    def start(self, timeout=100):
        if not self.container.is_running():
            self.container.start()

        running, _ = self.is_running()
        if running:
            return
        logger.debug('start %s', self)

        self._configure()
        self.container.client.system('/bin/ardb-server /etc/ardb.conf.used', id="{}.{}".format("storage_engine", self.name))

        # wait for storageEngine to start
        start = time.time()
        end = start + timeout
        is_running, _ = self.is_running()
        while not is_running and time.time() < end:
            time.sleep(1)
            is_running, _ = self.is_running()

        if not is_running:
            raise RuntimeError("storage server {} didn't started".format(self.name))

    def stop(self, timeout=30):
        if not self.container.is_running():
            return

        is_running, job = self.is_running()
        if not is_running:
            return

        logger.debug('stop %s', self)

        self.container.client.job.kill(job['cmd']['id'])

        # wait for StorageEngine to stop
        start = time.time()
        end = start + timeout
        is_running, _ = self.is_running()
        while is_running and time.time() < end:
            time.sleep(1)
            is_running, _ = self.is_running()

        if is_running:
            raise RuntimeError("storage server {} didn't stopped")

    def is_healthy(self):
        import redis
        client = redis.Redis(self.container.node.addr, self.port)
        key = "keytest"
        value = b"some test value"
        if not client.set(key, value):
            return False

        result = client.get(key)
        if result != value:
            return False
        client.delete(key)
        if client.exists(key):
            return False
        return True

    def is_running(self):
        try:
            if self.port not in self.container.node.freeports(self.port, 1):
                for job in self.container.client.job.list():
                    if 'name' in job['cmd']['arguments'] and job['cmd']['arguments']['name'] == '/bin/ardb-server':
                        return (True, job)
            return (False, None)
        except Exception as err:
            if str(err).find("invalid container id"):
                return (False, None)
            raise

    @property
    def ays(self):
        if self._ays is None:
            from JumpScale.sal.g8os.atyourservice.StorageCluster import storageEngineAys
            self._ays = storageEngineAys(self)
        return self._ays

    def __str__(self):
        return "storageEngine <{}>".format(self.name)

    def __repr__(self):
        return str(self)



"""
Auto-generated class for ClusterCreate
"""
from .EnumClusterCreateClusterType import EnumClusterCreateClusterType
from .EnumClusterCreateDriveType import EnumClusterCreateDriveType

from . import client_support


class ClusterCreate(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(clusterType, driveType, label, nodes, servers, k=None, m=None):
        """
        :type clusterType: EnumClusterCreateClusterType
        :type driveType: EnumClusterCreateDriveType
        :type k: int
        :type label: str
        :type m: int
        :type nodes: list[str]
        :type servers: int
        :rtype: ClusterCreate
        """

        return ClusterCreate(
            clusterType=clusterType,
            driveType=driveType,
            k=k,
            label=label,
            m=m,
            nodes=nodes,
            servers=servers,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'ClusterCreate'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'clusterType'
        val = data.get(property_name)
        if val is not None:
            datatypes = [EnumClusterCreateClusterType]
            try:
                self.clusterType = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'driveType'
        val = data.get(property_name)
        if val is not None:
            datatypes = [EnumClusterCreateDriveType]
            try:
                self.driveType = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'k'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.k = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

        property_name = 'label'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.label = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'm'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.m = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

        property_name = 'nodes'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.nodes = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'servers'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.servers = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



from zeroos.core0.client import Client
from .Disk import Disks, DiskType
from .Container import Containers
from .StoragePool import StoragePools
from .Network import Network
from .healthcheck import HealthCheck
from collections import namedtuple
from datetime import datetime
from io import BytesIO
import netaddr

Mount = namedtuple('Mount', ['device', 'mountpoint', 'fstype', 'options'])


class Node:
    """Represent a G8OS Server"""

    def __init__(self, addr, port=6379, password=None, timeout=120):
        # g8os client to talk to the node
        self._client = Client(host=addr, port=port, password=password, timeout=timeout)
        self._storageAddr = None
        self.addr = addr
        self.port = port
        self.disks = Disks(self)
        self.storagepools = StoragePools(self)
        self.containers = Containers(self)
        self.network = Network(self)
        self.healthcheck = HealthCheck(self)

    @classmethod
    def from_ays(cls, service, password=None, timeout=120):
        return cls(
            addr=service.model.data.redisAddr,
            port=service.model.data.redisPort,
            password=password,
            timeout=timeout
        )

    @property
    def client(self):
        return self._client

    @property
    def name(self):
        def get_nic_hwaddr(nics, name):
            for nic in nics:
                if nic['name'] == name:
                    return nic['hardwareaddr']

        defaultgwdev = self.client.bash("ip route | grep default | awk '{print $5}'").get().stdout.strip()
        nics = self.client.info.nic()
        macgwdev = None
        if defaultgwdev:
            macgwdev = get_nic_hwaddr(nics, defaultgwdev)
        if not macgwdev:
            raise AttributeError("name not find for node {}".format(self))
        return macgwdev.replace(":", '')

    @property
    def storageAddr(self):
        if not self._storageAddr:
            nic_data = self.client.info.nic()
            for nic in nic_data:
                if nic['name'] == 'backplane':
                    for ip in nic['addrs']:
                        network = netaddr.IPNetwork(ip['addr'])
                        if network.version == 4:
                            self._storageAddr = network.ip.format()
                            return self._storageAddr
            self._storageAddr = self.addr
        return self._storageAddr

    def get_nic_by_ip(self, addr):
        try:
            res = next(nic for nic in self._client.info.nic() if any(addr == a['addr'].split('/')[0] for a in nic['addrs']))
            return res
        except StopIteration:
            return None

    def _eligible_fscache_disk(self, disks):
        """
        return the first disk that is eligible to be used as filesystem cache
        First try to find a SSH disk, otherwise return a HDD
        """
        priorities = [DiskType.ssd, DiskType.hdd, DiskType.nvme]
        eligible = {t: [] for t in priorities}
        # Pick up the first ssd
        usedisks = []
        for pool in (self._client.btrfs.list() or []):
            for device in pool['devices']:
                usedisks.append(device['path'])
        for disk in disks[::-1]:
            if disk.devicename in usedisks or len(disk.partitions) > 0:
                continue
            if disk.type in priorities:
                eligible[disk.type].append(disk)
        # pick up the first disk according to priorities
        for t in priorities:
            if eligible[t]:
                return eligible[t][0]
        else:
            raise RuntimeError("cannot find eligible disks for the fs cache")

    def _mount_fscache(self, storagepool):
        """
        mount the fscache storage pool and copy the content of the in memmory fs inside
        """
        mountedpaths = [mount.mountpoint for mount in self.list_mounts()]
        containerpath = '/var/cache/containers'
        if containerpath not in mountedpaths:
            if storagepool.exists('containercache'):
                storagepool.get('containercache').delete()
            fs = storagepool.create('containercache')
            self.client.disk.mount(storagepool.devicename, containerpath, ['subvol={}'.format(fs.subvolume)])
        logpath = '/var/log'
        if logpath not in mountedpaths:
            # logs is empty filesystem which we create a snapshot on to store logs of current boot
            snapname = '{:%Y-%m-%d-%H-%M}'.format(datetime.now())
            fs = storagepool.get('logs')
            snapshot = fs.create(snapname)
            self.client.bash('mkdir /tmp/log && mv /var/log/* /tmp/log/')
            self.client.disk.mount(storagepool.devicename, logpath, ['subvol={}'.format(snapshot.subvolume)])
            self.client.bash('mv /tmp/log/* /var/log/').get()
            self.client.logger.reopen()
            # startup syslogd and klogd
            self.client.system('syslogd -n -O /var/log/messages')
            self.client.system('klogd -n')

    def freeports(self, baseport=2000, nrports=3):
        ports = self.client.info.port()
        usedports = set()
        for portInfo in ports:
            if portInfo['network'] != "tcp":
                continue
            usedports.add(portInfo['port'])

        freeports = []
        while True:
            if baseport not in usedports:
                freeports.append(baseport)
                if len(freeports) >= nrports:
                    return freeports
            baseport += 1

    def find_persistance(self, name='fscache'):
        fscache_sp = None
        for sp in self.storagepools.list():
            if sp.name == name:
                fscache_sp = sp
                break
        return fscache_sp

    def ensure_persistance(self, name='fscache'):
        """
        look for a disk not used,
        create a partition and mount it to be used as cache for the g8ufs
        set the label `fs_cache` to the partition
        """
        disks = self.disks.list()
        if len(disks) <= 0:
            # if no disks, we can't do anything
            return

        # check if there is already a storage pool with the fs_cache label
        fscache_sp = self.find_persistance(name)

        # create the storage pool if we don't have one yet
        if fscache_sp is None:
            disk = self._eligible_fscache_disk(disks)
            fscache_sp = self.storagepools.create(name, devices=[disk.devicename], metadata_profile='single', data_profile='single', overwrite=True)
        fscache_sp.mount()
        try:
            fscache_sp.get('logs')
        except ValueError:
            fscache_sp.create('logs')

        # mount the storage pool
        self._mount_fscache(fscache_sp)
        return fscache_sp

    def download_content(self, remote):
        buff = BytesIO()
        self.client.filesystem.download(remote, buff)
        return buff.getvalue().decode()

    def upload_content(self, remote, content):
        if isinstance(content, str):
            content = content.encode('utf8')
        bytes = BytesIO(content)
        self.client.filesystem.upload(remote, bytes)

    def wipedisks(self):
        print('Wiping node {hostname}'.format(**self.client.info.os()))
        mounteddevices = {mount['device']: mount for mount in self.client.info.disk()}

        def getmountpoint(device):
            for mounteddevice, mount in mounteddevices.items():
                if mounteddevice.startswith(device):
                    return mount

        jobs = []
        for disk in self.client.disk.list()['blockdevices']:
            devicename = '/dev/{}'.format(disk['kname'])
            if disk['tran'] == 'usb':
                print('   * Not wiping usb {kname} {model}'.format(**disk))
                continue
            mount = getmountpoint(devicename)
            if not mount:
                print('   * Wiping disk {kname}'.format(**disk))
                jobs.append(self.client.system('dd if=/dev/zero of={} bs=1M count=50'.format(devicename)))
            else:
                print('   * Not wiping {device} mounted at {mountpoint}'.format(device=devicename, mountpoint=mount['mountpoint']))

        # wait for wiping to complete
        for job in jobs:
            job.get()

    def list_mounts(self):
        allmounts = []
        for mount in self.client.info.disk():
            allmounts.append(Mount(mount['device'],
                                   mount['mountpoint'],
                                   mount['fstype'],
                                   mount['opts']))
        return allmounts

    def __str__(self):
        return "Node <{host}:{port}>".format(
            host=self.addr,
            port=self.port,
        )

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        a = "{}:{}".format(self.addr, self.port)
        b = "{}:{}".format(other.addr, other.port)
        return a == b

    def __hash__(self):
        return hash((self.addr, self.port))



import ssl
import json
import aioredis
import sys
import uuid
import time
import logging
import asyncio


logger = logging.getLogger('g8core')


class Response:
    def __init__(self, client, id):
        self._client = client
        self._id = id
        self._queue = 'result:{}'.format(id)

    async def exists(self):
        r = self._client._redis
        flag = '{}:flag'.format(self._queue)
        key_exists = await r.connection.execute('LKEYEXISTS', flag)
        return bool(key_exists)

    async def get(self, timeout=None):
        if timeout is None:
            timeout = self._client.timeout
        r = self._client._redis
        start = time.time()
        maxwait = timeout
        while maxwait > 0:
            job_exists = await self.exists()
            if not job_exists:
                raise RuntimeError("Job not found: %s" % self.id)
            v = await r.brpoplpush(self._queue, self._queue, min(maxwait, 10))
            if v is not None:
                return json.loads(v.decode())
            logger.debug('%s still waiting (%ss)', self._id, int(time.time() - start))
            maxwait -= 10
        raise TimeoutError()


class Pubsub:
    def __init__(self, loop, host, port=6379, password="", db=0, ctx=None, timeout=None, testConnectionAttempts=3, callback=None):

        socket_timeout = (timeout + 5) if timeout else 15

        self.testConnectionAttempts = testConnectionAttempts

        self._redis = None
        self.host = host
        self.port = port
        self.password = password
        self.db = db
        if ctx is None:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        self.ssl = ctx
        self.timeout = socket_timeout
        self.loop = loop

        async def default_callback(job_id, level, line, meta):
            w = sys.stdout if level == 1 else sys.stderr
            w.write(line)
            w.write('\n')

        self.callback = callback or default_callback
        if not callable(self.callback):
            raise Exception('callback must be callable')

    async def get(self):
        if self._redis is not None:
            return self._redis

        self._redis = await aioredis.create_redis((self.host, self.port),
                                                  loop=self.loop,
                                                  password=self.password,
                                                  db=self.db,
                                                  ssl=self.ssl,
                                                  timeout=self.timeout)
        return self._redis

    async def global_stream(self, queue, timeout=120):
        if self._redis.connection.closed:
            self._redis = await self.get()

        data = await asyncio.wait_for(self._redis.blpop(queue, timeout=timeout), timeout=timeout)

        _, body = data
        payload = json.loads(body.decode())
        message = payload['message']
        line = message['message']
        meta = message['meta']
        job_id = payload['command']
        await self.callback(job_id, meta >> 16, line, meta & 0xff)

    async def raw(self, command, arguments, queue=None, max_time=None, stream=False, tags=None, id=None):
        if not id:
            id = str(uuid.uuid4())

        payload = {
            'id': id,
            'command': command,
            'arguments': arguments,
            'queue': queue,
            'max_time': max_time,
            'stream': stream,
            'tags': tags,
        }

        self._redis = await self.get()
        flag = 'result:{}:flag'.format(id)

        await self._redis.rpush('core:default', json.dumps(payload))
        if await self._redis.brpoplpush(flag, flag, 10) is None:
            raise TimeoutError('failed to queue job {}'.format(id))
        logger.debug('%s >> g8core.%s(%s)', id, command, ', '.join(("%s=%s" % (k, v) for k, v in arguments.items())))

        return Response(self, id)

    async def sync(self, command, args):
        response = await self.raw(command, args)
        result = await response.get()
        if result["state"] != 'SUCCESS':
            raise RuntimeError('invalid response: %s' % result["state"])

        return json.loads(result["data"])

    async def ping(self):
        response = await self.sync('core.ping', {})
        return response

    async def subscribe(self, queue=None):
        response = await self.sync('logger.subscribe', {'queue': queue})
        return response



import re
from ..healthcheck import HealthCheckRun

descr = """
Monitors if disk usage is too high
"""


class DiskUsage(HealthCheckRun):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('disk-usage', 'Disk Usage Check', 'Hardware', resource)
        self.node = node

    def run(self):
        fs = self.node.client.btrfs.list()
        disks = {d['path']: d for f in fs for d in f['devices']}
        for path, disk in disks.items():
            usage = 100.0 * disk['used'] / disk['size']
            if usage > 95:
                self.add_message('{}_usage'.format(path), 'ERROR', "Disk usage of {} is {:.2%}".format(path, usage / 100))
            elif usage > 90:
                self.add_message('{}_usage'.format(path), 'WARNING', "Disk usage of {} is {:.2%}".format(path, usage / 100))
            else:
                self.add_message('{}_usage'.format(path), 'OK', "Disk usage of {} is {:.2%}".format(path, usage / 100))



from io import BytesIO
import logging
import yaml

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class EtcdCluster:
    """etced server"""

    def __init__(self, name, dialstrings):
        self.name = name
        self.dialstrings = dialstrings
        self._ays = None

    @classmethod
    def from_ays(cls, service, password=None):
        logger.debug("create storageEngine from service (%s)", service)

        dialstrings = set()
        for etcd_service in service.producers.get('etcd', []):
            etcd = ETCD.from_ays(etcd_service, password)
            dialstrings.add(etcd.clientBind)

        return cls(
            name=service.name,
            dialstrings=",".join(dialstrings),
        )


class ETCD:
    """etced server"""

    def __init__(self, name, container, serverBind, clientBind, peers, data_dir='/mnt/data'):
        self.name = name
        self.container = container
        self.serverBind = serverBind
        self.clientBind = clientBind
        self.data_dir = data_dir
        self.peers = ",".join(peers)
        self._ays = None

    @classmethod
    def from_ays(cls, service, password=None):
        logger.debug("create storageEngine from service (%s)", service)
        from .Container import Container
        container = Container.from_ays(service.parent, password)

        return cls(
            name=service.name,
            container=container,
            serverBind=service.model.data.serverBind,
            clientBind=service.model.data.clientBind,
            data_dir=service.model.data.homeDir,
            peers=service.model.data.peers,
        )

    def start(self):
        configpath = "/etc/etcd_{}.config".format(self.name)

        config = {
            "name": self.name,
            "initial-advertise-peer-urls": "http://{}".format(self.serverBind),
            "listen-peer-urls": "http://{}".format(self.serverBind),
            "listen-client-urls": "http://{}".format(self.clientBind),
            "advertise-client-urls": "http://{}".format(self.clientBind),
            "initial-cluster": self.peers,
            "data-dir": self.data_dir,
            "initial-cluster-state": "new"
        }
        yamlconfig = yaml.safe_dump(config, default_flow_style=False)
        configstream = BytesIO(yamlconfig.encode('utf8'))
        configstream.seek(0)
        self.container.client.filesystem.upload(configpath, configstream)
        cmd = '/bin/etcd --config-file %s' % configpath
        self.container.client.system(cmd, id="etcd.{}".format(self.name))
        if not self.container.is_port_listening(int(self.serverBind.split(":")[1])):
            raise RuntimeError('Failed to start etcd server: {}'.format(self.name))

    def put(self, key, value):
        if value.startswith("-"):
            value = "-- %s" % value
        if key.startswith("-"):
            key = "-- %s" % key
        cmd = '/bin/etcdctl \
          --endpoints {etcd} \
          put {key} "{value}"'.format(etcd=self.clientBind, key=key, value=value)
        return self.container.client.system(cmd, env={"ETCDCTL_API": "3"}).get()



import signal
import time

from zeroos.orchestrator.sal import templates
from js9 import j


class InfluxDB:
    def __init__(self, container, ip, port):
        self.container = container
        self.ip = ip
        self.port = port

    def apply_config(self):
        influx_conf = templates.render('influxdb.conf', ip=self.ip, port=self.port)
        self.container.upload_content('/etc/influxdb/influxdb.conf', influx_conf)

    def is_running(self):
        for process in self.container.client.process.list():
            if 'influxd' in process['cmdline']:
                try:
                    self.list_databases()
                except:
                    return False, process['pid']
                else:
                    return True, process['pid']
        return False, None

    def stop(self, timeout=30):
        is_running, pid = self.is_running()
        if not is_running:
            return

        self.container.client.process.kill(pid, signal.SIGTERM)
        start = time.time()
        end = start + timeout
        is_running, _ = self.is_running()
        while is_running and time.time() < end:
            time.sleep(1)
            is_running, _ = self.is_running()

        if is_running:
            raise RuntimeError('Failed to stop influxd.')

        if self.container.node.client.nft.rule_exists(self.port):
            self.container.node.client.nft.drop_port(self.port)

    def start(self, timeout=30):
        is_running, _ = self.is_running()
        if is_running:
            return

        self.apply_config()

        if not self.container.node.client.nft.rule_exists(self.port):
            self.container.node.client.nft.open_port(self.port)

        self.container.client.system('influxd')
        time.sleep(1)

        start = time.time()
        end = start + timeout
        is_running, _ = self.is_running()
        while not is_running and time.time() < end:
            time.sleep(1)
            is_running, _ = self.is_running()

        if not is_running:
            if self.container.node.client.nft.rule_exists(self.port):
                self.container.node.client.nft.drop_port(self.port)
            raise RuntimeError('Failed to start influxd.')

    def list_databases(self):
        client = j.clients.influxdb.get(self.ip, port=self.port)
        return client.get_list_database()

    def create_databases(self, databases):
        client = j.clients.influxdb.get(self.ip, port=self.port)

        for database in databases:
            client.create_database(database)

    def drop_databases(self, databases):
        client = j.clients.influxdb.get(self.ip, port=self.port)

        for database in databases:
            client.drop_database(database)



import re
from ..healthcheck import HealthCheckRun

descr = """
Monitors if a network bond (if there is one) has both (or more) interfaces properly active.
"""


class NetworkStability(HealthCheckRun):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('networkstability', 'Network Stability Check', 'Network',resource)
        self.node = node

    def run(self, nodes):
        nic = self.node.get_nic_by_ip(self.node.addr)
        if nic is None:
            raise LookupError("Couldn't get the management nic")
        jobs = []
        for node in nodes:
            other_nic = node.get_nic_by_ip(node.addr)
            if other_nic is not None:
                if nic['mtu'] != other_nic['mtu']:
                    self.add_message('{}_mtu'.format(node.name), 'ERROR', 'The management interface has mtu {} which is different than node {} which is {}'.format(nic['mtu'], node.name, other_nic['mtu']))
                else:
                    self.add_message('{}_mtu'.format(node.name), 'OK', 'The management interface has mtu {} is the same as node {}'.format(nic['mtu'], node.name, other_nic['mtu']))
            else:
                self.add_message('{}_mtu'.format(node.name), 'ERROR', "Couldn't get the management nic for node {}".format(node.name))
            jobs.append(self.node.client.system('ping -I {} -c 10 -W 1 -q {}'.format(self.node.addr, node.addr), max_time=20))
        for node, job in zip(nodes, jobs):
            res = job.get().stdout.split('\n')
            perc = 100 - int(res[2].split(',')[-1].strip().split()[0][:-1])
            if perc < 70:
                self.add_message('{}_ping_perc'.format(node.name), 'ERROR', "Can reach node {} with percentage {}".format(node.name, perc))
            elif perc < 90:
                self.add_message('{}_ping_perc'.format(node.name), 'WARNING', "Can reach node {} with percentage {}".format(node.name, perc))
            else:
                self.add_message('{}_ping_perc'.format(node.name), 'OK', "Can reach node {} with percentage {}".format(node.name, perc))
            if perc == 0:
                self.add_message('{}_ping_rt'.format(node.name), 'ERROR', "Can't reach node {}".format(node.name))
            else:
                rt = float(res[3].split('/')[3])
                if rt > 200:
                    self.add_message('{}_ping_rt'.format(node.name), 'ERROR', "Round-trip time to node {} is {}".format(node.name, rt))
                elif rt > 10:
                    self.add_message('{}_ping_rt'.format(node.name), 'WARNING', "Round-trip time to node {} is {}".format(node.name, rt))
                else:
                    self.add_message('{}_ping_rt'.format(node.name), 'OK', "Round-trip time to node {} is {}".format(node.name, rt))



"""
Auto-generated class for CPUStats
"""

from . import client_support


class CPUStats(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(guestNice, idle, ioWait, irq, nice, softIrq, steal, stolen, system, user):
        """
        :type guestNice: float
        :type idle: float
        :type ioWait: float
        :type irq: float
        :type nice: float
        :type softIrq: float
        :type steal: float
        :type stolen: float
        :type system: float
        :type user: float
        :rtype: CPUStats
        """

        return CPUStats(
            guestNice=guestNice,
            idle=idle,
            ioWait=ioWait,
            irq=irq,
            nice=nice,
            softIrq=softIrq,
            steal=steal,
            stolen=stolen,
            system=system,
            user=user,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'CPUStats'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'guestNice'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.guestNice = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'idle'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.idle = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'ioWait'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.ioWait = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'irq'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.irq = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'nice'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.nice = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'softIrq'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.softIrq = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'steal'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.steal = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'stolen'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.stolen = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'system'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.system = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'user'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.user = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



from ..healthcheck import HealthCheckRun
from js9 import j

descr = """
Clean up ssh deamons and tcp services from migration 
"""


class SSHCleanup(HealthCheckRun):
    def __init__(self, node, job):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('ssh-cleanup', 'SSH Cleanup', 'System Load', resource)
        self.node = node
        self.service = job.service
        self.job = job

    def run(self):
        status = 'OK'
        text = 'Migration Cleanup Succesful'
        finished = []
        try:
            for job in self.service.aysrepo.jobsList():
                job_dict = job.to_dict()
                if job_dict['actionName'] == 'processChange' and job_dict['actorName'] == 'vm':
                    if job_dict['state'] == 'running':
                        continue
                    vm = self.service.aysrepo.serviceGet(instance=job_dict['serviceName'], role=job_dict['actorName'])
                    finished.append("ssh.config_%s" % vm.name)
            for proc in self.node.client.process.list():
                for partial in finished:
                    if partial not in proc['cmdline']:
                        continue
                    config_file = proc['cmdline'].split()[-1]
                    port = config_file.split('_')[-1]
                    
                    self.node.client.process.kill(proc['pid'])
                    tcp_name = "tcp_%s_%s" % (self.node.name, port)
                    tcp_service = self.service.aysrepo.serviceGet(role='tcp', instance=tcp_name)
                    j.tools.async.wrappers.sync(tcp_service.executeAction("drop"), context=self.job.context)
                    tcp_service.delete()
                    if self.node.client.filesystem.exists('/tmp'):
                        self.node.client.filesystem.remove(config_file)

        except Exception as e:
            text = "Error happened, Can not clean ssh process "
            status = "ERROR"

        self.add_message(self.id, status, text)


import jinja2
import os


def get(templatename):
    templatepath = os.path.dirname(__file__)
    loader = jinja2.FileSystemLoader(templatepath)
    env = jinja2.Environment(loader=loader)
    return env.get_template(templatename)


def render(templatename, **kwargs):
    env = get(templatename)
    return env.render(**kwargs) + '\n'



class GraphsService:
    def __init__(self, client):
        self.client = client



    def DeleteDashboard(self, dashboardname, graphid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete a dashboard
        It is method for DELETE /graphs/{graphid}/dashboards/{dashboardname}
        """
        uri = self.client.base_url + "/graphs/"+graphid+"/dashboards/"+dashboardname
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetDashboard(self, dashboardname, graphid, headers=None, query_params=None, content_type="application/json"):
        """
        Get dashboard
        It is method for GET /graphs/{graphid}/dashboards/{dashboardname}
        """
        uri = self.client.base_url + "/graphs/"+graphid+"/dashboards/"+dashboardname
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListDashboards(self, graphid, headers=None, query_params=None, content_type="application/json"):
        """
        List dashboards
        It is method for GET /graphs/{graphid}/dashboards
        """
        uri = self.client.base_url + "/graphs/"+graphid+"/dashboards"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateDashboard(self, data, graphid, headers=None, query_params=None, content_type="application/json"):
        """
        Create Dashboard
        It is method for POST /graphs/{graphid}/dashboards
        """
        uri = self.client.base_url + "/graphs/"+graphid+"/dashboards"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetGraph(self, graphid, headers=None, query_params=None, content_type="application/json"):
        """
        Get a graph
        It is method for GET /graphs/{graphid}
        """
        uri = self.client.base_url + "/graphs/"+graphid
        return self.client.get(uri, None, headers, query_params, content_type)


    def UpdateGraph(self, data, graphid, headers=None, query_params=None, content_type="application/json"):
        """
        Update Graph
        It is method for PUT /graphs/{graphid}
        """
        uri = self.client.base_url + "/graphs/"+graphid
        return self.client.put(uri, data, headers, query_params, content_type)


    def ListGraphs(self, headers=None, query_params=None, content_type="application/json"):
        """
        List all graphs
        It is method for GET /graphs
        """
        uri = self.client.base_url + "/graphs"
        return self.client.get(uri, None, headers, query_params, content_type)
